@@ -5,13 +5,11 @@ lws_client_connect_2(struct lws *wsi)
 {
 #ifdef LWS_USE_IPV6
 	struct sockaddr_in6 server_addr6;
-	struct sockaddr_in6 client_addr6;
 	struct addrinfo hints, *result;
 #endif
 	struct lws_context *context = wsi->context;
 	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
 	struct sockaddr_in server_addr4;
-	struct sockaddr_in client_addr4;
 	struct lws_pollfd pfd;
 	struct sockaddr *v;
 	int n, plen = 0;
@@ -21,38 +19,38 @@ lws_client_connect_2(struct lws *wsi)
 
 	/* proxy? */
 
-	if (context->http_proxy_port) {
+	if (wsi->vhost->http_proxy_port) {
 		plen = sprintf((char *)pt->serv_buf,
 			"CONNECT %s:%u HTTP/1.0\x0d\x0a"
 			"User-agent: libwebsockets\x0d\x0a",
 			lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_PEER_ADDRESS),
-			wsi->u.hdr.ah->c_port);
+			wsi->u.hdr.c_port);
 
-		if (context->proxy_basic_auth_token[0])
+		if (wsi->vhost->proxy_basic_auth_token[0])
 			plen += sprintf((char *)pt->serv_buf + plen,
 					"Proxy-authorization: basic %s\x0d\x0a",
-					context->proxy_basic_auth_token);
+					wsi->vhost->proxy_basic_auth_token);
 
 		plen += sprintf((char *)pt->serv_buf + plen, "\x0d\x0a");
-		ads = context->http_proxy_address;
+		ads = wsi->vhost->http_proxy_address;
 
 #ifdef LWS_USE_IPV6
 		if (LWS_IPV6_ENABLED(context)) {
 			memset(&server_addr6, 0, sizeof(struct sockaddr_in6));
-			server_addr6.sin6_port = htons(context->http_proxy_port);
+			server_addr6.sin6_port = htons(wsi->vhost->http_proxy_port);
 		} else
 #endif
-			server_addr4.sin_port = htons(context->http_proxy_port);
+			server_addr4.sin_port = htons(wsi->vhost->http_proxy_port);
 
 	} else {
 		ads = lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_PEER_ADDRESS);
 #ifdef LWS_USE_IPV6
 		if (LWS_IPV6_ENABLED(context)) {
 			memset(&server_addr6, 0, sizeof(struct sockaddr_in6));
-			server_addr6.sin6_port = htons(wsi->u.hdr.ah->c_port);
+			server_addr6.sin6_port = htons(wsi->u.hdr.c_port);
 		} else
 #endif
-			server_addr4.sin_port = htons(wsi->u.hdr.ah->c_port);
+			server_addr4.sin_port = htons(wsi->u.hdr.c_port);
 	}
 
 	/*
@@ -114,8 +112,10 @@ lws_client_connect_2(struct lws *wsi)
 		ai.ai_socktype = SOCK_STREAM;
 		ai.ai_flags = AI_CANONNAME;
 
-		if (getaddrinfo(ads, NULL, &ai, &result))
+		if (getaddrinfo(ads, NULL, &ai, &result)) {
+			lwsl_err("getaddrinfo failed\n");
 			goto oom4;
+		}
 
 		res = result;
 		while (!p && res) {
@@ -129,6 +129,7 @@ lws_client_connect_2(struct lws *wsi)
 		}
 
 		if (!p) {
+			lwsl_err("Couldn't identify address\n");
 			freeaddrinfo(result);
 			goto oom4;
 		}
@@ -153,7 +154,7 @@ lws_client_connect_2(struct lws *wsi)
 			goto oom4;
 		}
 
-		if (lws_plat_set_socket_options(context, wsi->sock)) {
+		if (lws_plat_set_socket_options(wsi->vhost, wsi->sock)) {
 			lwsl_err("Failed to set wsi socket options\n");
 			compatible_close(wsi->sock);
 			goto oom4;
@@ -162,6 +163,7 @@ lws_client_connect_2(struct lws *wsi)
 		wsi->mode = LWSCM_WSCL_WAITING_CONNECT;
 
 		lws_libev_accept(wsi, wsi->sock);
+		lws_libuv_accept(wsi, wsi->sock);
 		if (insert_wsi_socket_into_fds(context, wsi)) {
 			compatible_close(wsi->sock);
 			goto oom4;
@@ -171,40 +173,19 @@ lws_client_connect_2(struct lws *wsi)
 		 * past here, we can't simply free the structs as error
 		 * handling as oom4 does.  We have to run the whole close flow.
 		 */
+
+		if (!wsi->protocol)
+			wsi->protocol = &wsi->vhost->protocols[0];
+
 		wsi->protocol->callback(wsi, LWS_CALLBACK_WSI_CREATE,
 					wsi->user_space, NULL, 0);
-		lws_set_timeout(wsi,
-			PENDING_TIMEOUT_AWAITING_CONNECT_RESPONSE,
-							      AWAITING_TIMEOUT);
-#ifdef LWS_USE_IPV6
-		if (LWS_IPV6_ENABLED(context)) {
-			v = (struct sockaddr *)&client_addr6;
-			n = sizeof(client_addr6);
-			bzero((char *)v, n);
-			client_addr6.sin6_family = AF_INET6;
-		} else
-#endif
-		{
-			v = (struct sockaddr *)&client_addr4;
-			n = sizeof(client_addr4);
-			bzero((char *)v, n);
-			client_addr4.sin_family = AF_INET;
-		}
 
-		if (context->iface) {
-			if (interface_to_sa(context, context->iface,
-					(struct sockaddr_in *)v, n) < 0) {
-				lwsl_err("Unable to find interface %s\n",
-								context->iface);
-				goto failed;
-			}
+		lws_set_timeout(wsi, PENDING_TIMEOUT_AWAITING_CONNECT_RESPONSE,
+				AWAITING_TIMEOUT);
 
-			if (bind(wsi->sock, v, n) < 0) {
-				lwsl_err("Error binding to interface %s",
-								context->iface);
-				goto failed;
-			}
-		}
+		n = lws_socket_bind(wsi->vhost, wsi->sock, 0, wsi->vhost->iface);
+		if (n < 0)
+			goto failed;
 	}
 
 #ifdef LWS_USE_IPV6
@@ -248,7 +229,7 @@ lws_client_connect_2(struct lws *wsi)
 
 	/* we are connected to server, or proxy */
 
-	if (context->http_proxy_port) {
+	if (wsi->vhost->http_proxy_port) {
 
 		/*
 		 * OK from now on we talk via the proxy, so connect to that
@@ -257,9 +238,9 @@ lws_client_connect_2(struct lws *wsi)
 		 * leaving old string/frag there but unreferenced)
 		 */
 		if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_PEER_ADDRESS,
-					  context->http_proxy_address))
+					  wsi->vhost->http_proxy_address))
 			goto failed;
-		wsi->u.hdr.ah->c_port = context->http_proxy_port;
+		wsi->u.hdr.c_port = wsi->vhost->http_proxy_port;
 
 		n = send(wsi->sock, (char *)pt->serv_buf, plen,
 			 MSG_NOSIGNAL);
@@ -291,6 +272,7 @@ lws_client_connect_2(struct lws *wsi)
 
 	wsi->mode = LWSCM_WSCL_ISSUE_HANDSHAKE;
 	pfd.fd = wsi->sock;
+	pfd.events = LWS_POLLIN;
 	pfd.revents = LWS_POLLIN;
 
 	n = lws_service_fd(context, &pfd);
@@ -304,7 +286,15 @@ lws_client_connect_2(struct lws *wsi)
 oom4:
 	/* we're closing, losing some rx is OK */
 	wsi->u.hdr.ah->rxpos = wsi->u.hdr.ah->rxlen;
-	lws_header_table_detach(wsi);
+	//lwsl_err("%d\n", wsi->mode);
+	if (wsi->mode == LWSCM_HTTP_CLIENT)
+		wsi->vhost->protocols[0].callback(wsi,
+			LWS_CALLBACK_CLIENT_CONNECTION_ERROR,
+			wsi->user_space, NULL, 0);
+	/* take care that we might be inserted in fds already */
+	if (wsi->position_in_fds_table != -1)
+		goto failed;
+	lws_header_table_detach(wsi, 0);
 	lws_free(wsi);
 
 	return NULL;
@@ -359,11 +349,120 @@ lws_client_reset(struct lws *wsi, int ssl, const char *address, int port, const
 	wsi->state = LWSS_CLIENT_UNCONNECTED;
 	wsi->protocol = NULL;
 	wsi->pending_timeout = NO_PENDING_TIMEOUT;
-	wsi->u.hdr.ah->c_port = port;
+	wsi->u.hdr.c_port = port;
 
 	return lws_client_connect_2(wsi);
 }
 
+#ifdef LWS_WITH_HTTP_PROXY
+static hubbub_error
+html_parser_cb(const hubbub_token *token, void *pw)
+{
+	struct lws_rewrite *r = (struct lws_rewrite *)pw;
+	char buf[1024], *start = buf + LWS_PRE, *p = start,
+	     *end = &buf[sizeof(buf) - 1];
+	size_t i;
+
+	switch (token->type) {
+	case HUBBUB_TOKEN_DOCTYPE:
+
+		p += snprintf(p, end - p, "<!DOCTYPE %.*s %s ",
+				(int) token->data.doctype.name.len,
+				token->data.doctype.name.ptr,
+				token->data.doctype.force_quirks ?
+						"(force-quirks) " : "");
+
+		if (token->data.doctype.public_missing)
+			printf("\tpublic: missing\n");
+		else
+			p += snprintf(p, end - p, "PUBLIC \"%.*s\"\n",
+				(int) token->data.doctype.public_id.len,
+				token->data.doctype.public_id.ptr);
+
+		if (token->data.doctype.system_missing)
+			printf("\tsystem: missing\n");
+		else
+			p += snprintf(p, end - p, " \"%.*s\">\n",
+				(int) token->data.doctype.system_id.len,
+				token->data.doctype.system_id.ptr);
+
+		break;
+	case HUBBUB_TOKEN_START_TAG:
+		p += snprintf(p, end - p, "<%.*s", (int)token->data.tag.name.len,
+				token->data.tag.name.ptr);
+
+/*				(token->data.tag.self_closing) ?
+						"(self-closing) " : "",
+				(token->data.tag.n_attributes > 0) ?
+						"attributes:" : "");
+*/
+		for (i = 0; i < token->data.tag.n_attributes; i++) {
+			if (!hstrcmp(&token->data.tag.attributes[i].name, "href", 4) ||
+			    !hstrcmp(&token->data.tag.attributes[i].name, "action", 6) ||
+			    !hstrcmp(&token->data.tag.attributes[i].name, "src", 3)) {
+				const char *pp = (const char *)token->data.tag.attributes[i].value.ptr;
+				int plen = (int) token->data.tag.attributes[i].value.len;
+
+				if (!hstrcmp(&token->data.tag.attributes[i].value,
+					     r->from, r->from_len)) {
+					pp += r->from_len;
+					plen -= r->from_len;
+				}
+				p += snprintf(p, end - p, " %.*s=\"%s/%.*s\"",
+				       (int) token->data.tag.attributes[i].name.len,
+				       token->data.tag.attributes[i].name.ptr,
+				       r->to, plen, pp);
+
+			} else
+
+				p += snprintf(p, end - p, " %.*s=\"%.*s\"",
+					(int) token->data.tag.attributes[i].name.len,
+					token->data.tag.attributes[i].name.ptr,
+					(int) token->data.tag.attributes[i].value.len,
+					token->data.tag.attributes[i].value.ptr);
+		}
+		p += snprintf(p, end - p, ">\n");
+		break;
+	case HUBBUB_TOKEN_END_TAG:
+		p += snprintf(p, end - p, "</%.*s", (int) token->data.tag.name.len,
+				token->data.tag.name.ptr);
+/*
+				(token->data.tag.self_closing) ?
+						"(self-closing) " : "",
+				(token->data.tag.n_attributes > 0) ?
+						"attributes:" : "");
+*/
+		for (i = 0; i < token->data.tag.n_attributes; i++) {
+			p += snprintf(p, end - p, " %.*s='%.*s'\n",
+				(int) token->data.tag.attributes[i].name.len,
+				token->data.tag.attributes[i].name.ptr,
+				(int) token->data.tag.attributes[i].value.len,
+				token->data.tag.attributes[i].value.ptr);
+		}
+		p += snprintf(p, end - p, ">\n");
+		break;
+	case HUBBUB_TOKEN_COMMENT:
+		p += snprintf(p, end - p, "<!-- %.*s -->\n",
+				(int) token->data.comment.len,
+				token->data.comment.ptr);
+		break;
+	case HUBBUB_TOKEN_CHARACTER:
+		p += snprintf(p, end - p, "%.*s", (int) token->data.character.len,
+				token->data.character.ptr);
+		break;
+	case HUBBUB_TOKEN_EOF:
+		p += snprintf(p, end - p, "\n");
+		break;
+	}
+
+	if (user_callback_handle_rxflow(r->wsi->protocol->callback,
+			r->wsi, LWS_CALLBACK_RECEIVE_CLIENT_HTTP_READ,
+			r->wsi->user_space, start, p - start))
+		return -1;
+
+	return HUBBUB_OK;
+}
+#endif
 /**
  * lws_client_connect_via_info() - Connect to another websocket server
  * @i:pointer to lws_client_connect_info struct
@@ -378,15 +477,24 @@ lws_client_connect_via_info(struct lws_client_connect_info *i)
 	struct lws *wsi;
 	int v = SPEC_LATEST_SUPPORTED;
 
+	if (i->context->requested_kill)
+		return NULL;
+
 	wsi = lws_zalloc(sizeof(struct lws));
 	if (wsi == NULL)
 		goto bail;
 
 	wsi->context = i->context;
+	/* assert the mode and union status (hdr) clearly */
+	lws_union_transition(wsi, LWSCM_HTTP_CLIENT);
 	wsi->sock = LWS_SOCK_INVALID;
 
-	/* -1 means just use latest supported */
+	/* 1) fill up the wsi with stuff from the connect_info as far as it
+	 * can go.  It's because not only is our connection async, we might
+	 * not even be able to get ahold of an ah at this point.
+	 */
 
+	/* -1 means just use latest supported */
 	if (i->ietf_version_or_minus_one != -1 && i->ietf_version_or_minus_one)
 		v = i->ietf_version_or_minus_one;
 
@@ -396,6 +504,22 @@ lws_client_connect_via_info(struct lws_client_connect_info *i)
 	wsi->protocol = NULL;
 	wsi->pending_timeout = NO_PENDING_TIMEOUT;
 	wsi->position_in_fds_table = -1;
+	wsi->u.hdr.c_port = i->port;
+	wsi->vhost = i->vhost;
+	if (!wsi->vhost)
+		wsi->vhost = i->context->vhost_list;
+
+	wsi->protocol = &wsi->vhost->protocols[0];
+	if (wsi && !wsi->user_space && i->userdata) {
+		wsi->user_space_externally_allocated = 1;
+		wsi->user_space = i->userdata;
+	} else
+		/* if we stay in http, we can assign the user space now,
+		 * otherwise do it after the protocol negotiated
+		 */
+		if (i->method)
+			if (lws_ensure_user_space(wsi))
+				goto bail;
 
 #ifdef LWS_OPENSSL_SUPPORT
 	wsi->use_ssl = i->ssl_connection;
@@ -406,42 +530,116 @@ lws_client_connect_via_info(struct lws_client_connect_info *i)
 	}
 #endif
 
-	if (lws_header_table_attach(wsi))
+	/* 2) stash the things from connect_info that we can't process without
+	 * an ah.  Because if no ah, we will go on the ah waiting list and
+	 * process those things later (after the connect_info and maybe the
+	 * things pointed to have gone out of scope.
+	 */
+
+	wsi->u.hdr.stash = lws_malloc(sizeof(*wsi->u.hdr.stash));
+	if (!wsi->u.hdr.stash) {
+		lwsl_err("%s: OOM\n", __func__);
 		goto bail;
+	}
+
+	wsi->u.hdr.stash->origin[0] = '\0';
+	wsi->u.hdr.stash->protocol[0] = '\0';
+	wsi->u.hdr.stash->method[0] = '\0';
+
+	strncpy(wsi->u.hdr.stash->address, i->address,
+		sizeof(wsi->u.hdr.stash->address) - 1);
+	strncpy(wsi->u.hdr.stash->path, i->path,
+		sizeof(wsi->u.hdr.stash->path) - 1);
+	strncpy(wsi->u.hdr.stash->host, i->host,
+		sizeof(wsi->u.hdr.stash->host) - 1);
+	if (i->origin)
+		strncpy(wsi->u.hdr.stash->origin, i->origin,
+			sizeof(wsi->u.hdr.stash->origin) - 1);
+	if (i->protocol)
+		strncpy(wsi->u.hdr.stash->protocol, i->protocol,
+			sizeof(wsi->u.hdr.stash->protocol) - 1);
+	if (i->method)
+		strncpy(wsi->u.hdr.stash->method, i->method,
+			sizeof(wsi->u.hdr.stash->method) - 1);
+
+	wsi->u.hdr.stash->address[sizeof(wsi->u.hdr.stash->address) - 1] = '\0';
+	wsi->u.hdr.stash->path[sizeof(wsi->u.hdr.stash->path) - 1] = '\0';
+	wsi->u.hdr.stash->host[sizeof(wsi->u.hdr.stash->host) - 1] = '\0';
+	wsi->u.hdr.stash->origin[sizeof(wsi->u.hdr.stash->origin) - 1] = '\0';
+	wsi->u.hdr.stash->protocol[sizeof(wsi->u.hdr.stash->protocol) - 1] = '\0';
+	wsi->u.hdr.stash->method[sizeof(wsi->u.hdr.stash->method) - 1] = '\0';
+
+	/* if we went on the waiting list, no probs just return the wsi
+	 * when we get the ah, now or later, he will call
+	 * lws_client_connect_via_info2() below.
+	 */
+	if (lws_header_table_attach(wsi, 0) < 0)
+		return NULL;
+
+	if (i->parent_wsi) {
+		lwsl_info("%s: created child %p of parent %p\n", __func__,
+				wsi, i->parent_wsi);
+		wsi->parent = i->parent_wsi;
+		wsi->sibling_list = i->parent_wsi->child_list;
+		i->parent_wsi->child_list = wsi;
+	}
+#ifdef LWS_WITH_HTTP_PROXY
+	if (i->uri_replace_to)
+		wsi->rw = lws_rewrite_create(wsi, html_parser_cb,
+					     i->uri_replace_from,
+					     i->uri_replace_to);
+#endif
+
+	return wsi;
+
+bail:
+	lws_free(wsi);
+
+	return NULL;
+}
+
+struct lws *
+lws_client_connect_via_info2(struct lws *wsi)
+{
+	struct client_info_stash *stash = wsi->u.hdr.stash;
+
+	if (!stash)
+		return wsi;
 
 	/*
 	 * we're not necessarily in a position to action these right away,
 	 * stash them... we only need during connect phase so u.hdr is fine
 	 */
-	wsi->u.hdr.ah->c_port = i->port;
-	if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_PEER_ADDRESS, i->address))
+	if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_PEER_ADDRESS,
+				  stash->address))
 		goto bail1;
 
 	/* these only need u.hdr lifetime as well */
 
-	if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_URI, i->path))
+	if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_URI, stash->path))
 		goto bail1;
 
-	if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_HOST, i->host))
+	if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_HOST, stash->host))
 		goto bail1;
 
-	if (i->origin)
-		if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_ORIGIN, i->origin))
+	if (stash->origin[0])
+		if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_ORIGIN,
+					  stash->origin))
 			goto bail1;
 	/*
 	 * this is a list of protocols we tell the server we're okay with
 	 * stash it for later when we compare server response with it
 	 */
-	if (i->protocol)
+	if (stash->protocol[0])
 		if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_SENT_PROTOCOLS,
-					  i->protocol))
+					  stash->protocol))
+			goto bail1;
+	if (stash->method[0])
+		if (lws_hdr_simple_create(wsi, _WSI_TOKEN_CLIENT_METHOD,
+					  stash->method))
 			goto bail1;
 
-	wsi->protocol = &i->context->protocols[0];
-	if (wsi && !wsi->user_space && i->userdata) {
-		wsi->user_space_externally_allocated = 1;
-		wsi->user_space = i->userdata;
-	}
+	lws_free_set_NULL(wsi->u.hdr.stash);
 
 	/*
 	 * Check with each extension if it is able to route and proxy this
@@ -450,9 +648,10 @@ lws_client_connect_via_info(struct lws_client_connect_info *i)
 	 * connection.
 	 */
 
-	if (lws_ext_cb_all_exts(i->context, wsi,
-			LWS_EXT_CB_CAN_PROXY_CLIENT_CONNECTION,
-						     (void *)i->address, i->port) > 0) {
+	if (lws_ext_cb_all_exts(wsi->context, wsi,
+				LWS_EXT_CB_CAN_PROXY_CLIENT_CONNECTION,
+				(void *)stash->address,
+				wsi->u.hdr.c_port) > 0) {
 		lwsl_client("lws_client_connect: ext handling conn\n");
 
 		lws_set_timeout(wsi,
@@ -463,17 +662,12 @@ lws_client_connect_via_info(struct lws_client_connect_info *i)
 		return wsi;
 	}
 	lwsl_client("lws_client_connect: direct conn\n");
-
 	wsi->context->count_wsi_allocated++;
 
 	return lws_client_connect_2(wsi);
 
 bail1:
-	/* we're closing, losing some rx is OK */
-	wsi->u.hdr.ah->rxpos = wsi->u.hdr.ah->rxlen;
-	lws_header_table_detach(wsi);
-bail:
-	lws_free(wsi);
+	lws_free_set_NULL(wsi->u.hdr.stash);
 
 	return NULL;
 }
