@@ -148,125 +148,15 @@ lws_client_socket_service(struct lws_context *context, struct lws *wsi,
 #ifdef LWS_OPENSSL_SUPPORT
 		/* we can retry this... just cook the SSL BIO the first time */
 
-		if (wsi->use_ssl && !wsi->ssl) {
-#if defined(CYASSL_SNI_HOST_NAME) || defined(WOLFSSL_SNI_HOST_NAME) || defined(SSL_CTRL_SET_TLSEXT_HOSTNAME)
-			const char *hostname = lws_hdr_simple_ptr(wsi,
-						_WSI_TOKEN_CLIENT_HOST);
-#endif
-
-			wsi->ssl = SSL_new(context->ssl_client_ctx);
-#ifndef USE_WOLFSSL
-			SSL_set_mode(wsi->ssl,
-					SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
-#endif
-			/*
-			 * use server name indication (SNI), if supported,
-			 * when establishing connection
-			 */
-#ifdef USE_WOLFSSL
-#ifdef USE_OLD_CYASSL
-#ifdef CYASSL_SNI_HOST_NAME
-			CyaSSL_UseSNI(wsi->ssl, CYASSL_SNI_HOST_NAME,
-				hostname, strlen(hostname));
-#endif
-#else
-#ifdef WOLFSSL_SNI_HOST_NAME
-			wolfSSL_UseSNI(wsi->ssl, WOLFSSL_SNI_HOST_NAME,
-				hostname, strlen(hostname));
-#endif
-#endif
-#else
-#ifdef SSL_CTRL_SET_TLSEXT_HOSTNAME
-			SSL_set_tlsext_host_name(wsi->ssl, hostname);
-#endif
-#endif
-
-#ifdef USE_WOLFSSL
-			/*
-			 * wolfSSL/CyaSSL does certificate verification differently
-			 * from OpenSSL.
-			 * If we should ignore the certificate, we need to set
-			 * this before SSL_new and SSL_connect is called.
-			 * Otherwise the connect will simply fail with error
-			 * code -155
-			 */
-#ifdef USE_OLD_CYASSL
-			if (wsi->use_ssl == 2)
-				CyaSSL_set_verify(wsi->ssl,
-							SSL_VERIFY_NONE, NULL);
-#else
-			if (wsi->use_ssl == 2)
-				wolfSSL_set_verify(wsi->ssl,
-							SSL_VERIFY_NONE, NULL);
-#endif
-#endif /* USE_WOLFSSL */
-
-			wsi->client_bio =
-				BIO_new_socket(wsi->sock, BIO_NOCLOSE);
-			SSL_set_bio(wsi->ssl, wsi->client_bio, wsi->client_bio);
-
-#ifdef USE_WOLFSSL
-#ifdef USE_OLD_CYASSL
-			CyaSSL_set_using_nonblock(wsi->ssl, 1);
-#else
-			wolfSSL_set_using_nonblock(wsi->ssl, 1);
-#endif
-#else
-			BIO_set_nbio(wsi->client_bio, 1); /* nonblocking */
-#endif
-
-			SSL_set_ex_data(wsi->ssl,
-					openssl_websocket_private_data_index,
-								       context);
-		}
+		if (wsi->use_ssl && !wsi->ssl)
+			lws_ssl_client_bio_create(wsi);
 
 		if (wsi->use_ssl) {
-			lws_latency_pre(context, wsi);
-			n = SSL_connect(wsi->ssl);
-			lws_latency(context, wsi,
-			  "SSL_connect LWSCM_WSCL_ISSUE_HANDSHAKE", n, n > 0);
-
-			if (n < 0) {
-				n = SSL_get_error(wsi->ssl, n);
-
-				if (n == SSL_ERROR_WANT_READ)
-					goto some_wait;
-
-				if (n == SSL_ERROR_WANT_WRITE) {
-					/*
-					 * wants us to retry connect due to
-					 * state of the underlying ssl layer...
-					 * but since it may be stalled on
-					 * blocked write, no incoming data may
-					 * arrive to trigger the retry.
-					 * Force (possibly many times if the SSL
-					 * state persists in returning the
-					 * condition code, but other sockets
-					 * are getting serviced inbetweentimes)
-					 * us to get called back when writable.
-					 */
-					lwsl_info("%s: WANT_WRITE... retrying\n", __func__);
-					lws_callback_on_writable(wsi);
-some_wait:
-					wsi->mode = LWSCM_WSCL_WAITING_SSL;
-
-					return 0; /* no error */
-				}
-				n = -1;
-			}
-
-			if (n <= 0) {
-				/*
-				 * retry if new data comes until we
-				 * run into the connection timeout or win
-				 */
-				n = ERR_get_error();
-				if (n != SSL_ERROR_NONE) {
-					lwsl_err("SSL connect error %lu: %s\n",
-						n, ERR_error_string(n, sb));
-					return 0;
-				}
-			}
+			n = lws_ssl_client_connect1(wsi);
+			if (!n)
+				return 0;
+			if (n < 0)
+				goto bail3;
 		} else
 			wsi->ssl = NULL;
 
@@ -275,78 +165,11 @@ some_wait:
 	case LWSCM_WSCL_WAITING_SSL:
 
 		if (wsi->use_ssl) {
-			if (wsi->mode == LWSCM_WSCL_WAITING_SSL) {
-				lws_latency_pre(context, wsi);
-				n = SSL_connect(wsi->ssl);
-				lws_latency(context, wsi,
-					    "SSL_connect LWSCM_WSCL_WAITING_SSL",
-					    n, n > 0);
-
-				if (n < 0) {
-					n = SSL_get_error(wsi->ssl, n);
-
-					if (n == SSL_ERROR_WANT_READ)
-						goto some_wait;
-
-					if (n == SSL_ERROR_WANT_WRITE) {
-						/*
-						 * wants us to retry connect due to
-						 * state of the underlying ssl layer...
-						 * but since it may be stalled on
-						 * blocked write, no incoming data may
-						 * arrive to trigger the retry.
-						 * Force (possibly many times if the SSL
-						 * state persists in returning the
-						 * condition code, but other sockets
-						 * are getting serviced inbetweentimes)
-						 * us to get called back when writable.
-						 */
-						lwsl_info("SSL_connect WANT_WRITE... retrying\n");
-						lws_callback_on_writable(wsi);
-
-						goto some_wait;
-					}
-					n = -1;
-				}
-
-				if (n <= 0) {
-					/*
-					 * retry if new data comes until we
-					 * run into the connection timeout or win
-					 */
-					n = ERR_get_error();
-					if (n != SSL_ERROR_NONE) {
-						lwsl_err("SSL connect error %lu: %s\n",
-							 n, ERR_error_string(n, sb));
-						return 0;
-					}
-				}
-			}
-
-			#ifndef USE_WOLFSSL
-			/*
-			 * See comment above about wolfSSL certificate
-			 * verification
-			 */
-			lws_latency_pre(context, wsi);
-			n = SSL_get_verify_result(wsi->ssl);
-			lws_latency(context, wsi,
-				"SSL_get_verify_result LWS_CONNMODE..HANDSHAKE",
-								      n, n > 0);
-
-			if (n != X509_V_OK) {
-				if ((n == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT ||
-				     n == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN) && wsi->use_ssl == 2) {
-					lwsl_notice("accepting self-signed certificate\n");
-				} else {
-					lwsl_err("server's cert didn't look good, X509_V_ERR = %d: %s\n",
-						 n, ERR_error_string(n, sb));
-					lws_close_free_wsi(wsi,
-						LWS_CLOSE_STATUS_NOSTATUS);
-					return 0;
-				}
-			}
-#endif /* USE_WOLFSSL */
+			n = lws_ssl_client_connect2(wsi);
+			if (!n)
+				return 0;
+			if (n < 0)
+				goto bail3;
 		} else
 			wsi->ssl = NULL;
 #endif
@@ -462,6 +285,9 @@ some_wait:
 
 bail3:
 		lwsl_info("closing conn at LWS_CONNMODE...SERVER_REPLY\n");
+		wsi->vhost->protocols[0].callback(wsi,
+			LWS_CALLBACK_CLIENT_CONNECTION_ERROR,
+					wsi->user_space, NULL, 0);
 		lws_close_free_wsi(wsi, LWS_CLOSE_STATUS_NOSTATUS);
 		return -1;
 
@@ -492,15 +318,61 @@ strtolower(char *s)
 	}
 }
 
+/**
+ * lws_http_transaction_completed() - wait for new http transaction or close
+ * @wsi:	websocket connection
+ *
+ *	Returns 1 if the HTTP connection must close now
+ *	Returns 0 and resets connection to wait for new HTTP header /
+ *	  transaction if possible
+ */
+
+int LWS_WARN_UNUSED_RESULT
+lws_http_transaction_completed_client(struct lws *wsi)
+{
+	lwsl_debug("%s: wsi %p\n", __func__, wsi);
+	/* if we can't go back to accept new headers, drop the connection */
+	if (wsi->u.http.connection_type != HTTP_CONNECTION_KEEP_ALIVE) {
+		lwsl_info("%s: %p: close connection\n", __func__, wsi);
+		return 1;
+	}
+
+	/* otherwise set ourselves up ready to go again */
+	wsi->state = LWSS_CLIENT_HTTP_ESTABLISHED;
+	wsi->mode = LWSCM_HTTP_CLIENT_ACCEPTED;
+	wsi->u.http.content_length = 0;
+	wsi->hdr_parsing_completed = 0;
+
+	/* He asked for it to stay alive indefinitely */
+	lws_set_timeout(wsi, NO_PENDING_TIMEOUT, 0);
+
+	/*
+	 * As client, nothing new is going to come until we ask for it
+	 * we can drop the ah, if any
+	 */
+	if (wsi->u.hdr.ah) {
+		wsi->u.hdr.ah->rxpos = wsi->u.hdr.ah->rxlen;
+		lws_header_table_detach(wsi, 0);
+	}
+
+	/* If we're (re)starting on headers, need other implied init */
+	wsi->u.hdr.ues = URIES_IDLE;
+
+	lwsl_info("%s: %p: keep-alive await new transaction\n", __func__, wsi);
+
+	return 0;
+}
+
 int
 lws_client_interpret_server_handshake(struct lws *wsi)
 {
 	int n, len, okay = 0, isErrorCodeReceived = 0, port = 0, ssl = 0;
-	struct lws_context *context = wsi->context;
 	int close_reason = LWS_CLOSE_STATUS_PROTOCOL_ERR;
 	const char *pc, *prot, *ads = NULL, *path;
+	struct allocated_headers *ah;
 	char *p;
 #ifndef LWS_NO_EXTENSIONS
+	struct lws_context *context = wsi->context;
 	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
 	char *sb = (char *)&pt->serv_buf[0];
 	const struct lws_ext_options *opts;
@@ -512,11 +384,41 @@ lws_client_interpret_server_handshake(struct lws *wsi)
 	void *v;
 #endif
 
+	if (!wsi->do_ws) {
+		/* we are being an http client...
+		 */
+		ah = wsi->u.hdr.ah;
+		lws_union_transition(wsi, LWSCM_HTTP_CLIENT_ACCEPTED);
+		wsi->state = LWSS_CLIENT_HTTP_ESTABLISHED;
+		wsi->u.http.ah = ah;
+	}
+
 	/*
 	 * well, what the server sent looked reasonable for syntax.
 	 * Now let's confirm it sent all the necessary headers
+	 *
+	 * http (non-ws) client will expect something like this
+	 *
+	 * HTTP/1.0.200
+	 * server:.libwebsockets
+	 * content-type:.text/html
+	 * content-length:.17703
+	 * set-cookie:.test=LWS_1456736240_336776_COOKIE;Max-Age=360000
+	 *
+	 *
+	 *
 	 */
+
+	wsi->u.http.connection_type = HTTP_CONNECTION_KEEP_ALIVE;
 	p = lws_hdr_simple_ptr(wsi, WSI_TOKEN_HTTP);
+	if (wsi->do_ws && !p) {
+		lwsl_info("no URI\n");
+		goto bail3;
+	}
+	if (!p) {
+		p = lws_hdr_simple_ptr(wsi, WSI_TOKEN_HTTP1_0);
+		wsi->u.http.connection_type = HTTP_CONNECTION_CLOSE;
+	}
 	if (!p) {
 		lwsl_info("no URI\n");
 		goto bail3;
@@ -539,6 +441,77 @@ lws_client_interpret_server_handshake(struct lws *wsi)
 		}
 		return 0;
 	}
+
+	if (!wsi->do_ws) {
+		if (n != 200) {
+			lwsl_notice("Connection failed with code %d", n);
+			goto bail2;
+		}
+
+#ifdef LWS_WITH_HTTP_PROXY
+		wsi->perform_rewrite = 0;
+		if (lws_hdr_total_length(wsi, WSI_TOKEN_HTTP_CONTENT_TYPE)) {
+			if (!strncmp(lws_hdr_simple_ptr(wsi, WSI_TOKEN_HTTP_CONTENT_TYPE),
+				     "text/html", 9))
+				wsi->perform_rewrite = 1;
+		}
+#endif
+
+		/* allocate the per-connection user memory (if any) */
+		if (lws_ensure_user_space(wsi)) {
+			lwsl_err("Problem allocating wsi user mem\n");
+			goto bail2;
+		}
+
+		/* he may choose to send us stuff in chunked transfer-coding */
+		wsi->chunked = 0;
+		wsi->chunk_remaining = 0; /* ie, next thing is chunk size */
+		if (lws_hdr_total_length(wsi, WSI_TOKEN_HTTP_TRANSFER_ENCODING)) {
+			wsi->chunked = !strcmp(lws_hdr_simple_ptr(wsi,
+					       WSI_TOKEN_HTTP_TRANSFER_ENCODING),
+					"chunked");
+			/* first thing is hex, after payload there is crlf */
+			wsi->chunk_parser = ELCP_HEX;
+		}
+
+		if (lws_hdr_total_length(wsi, WSI_TOKEN_HTTP_CONTENT_LENGTH)) {
+			wsi->u.http.content_length =
+					atoi(lws_hdr_simple_ptr(wsi,
+						WSI_TOKEN_HTTP_CONTENT_LENGTH));
+			lwsl_notice("%s: incoming content length %d\n", __func__,
+					wsi->u.http.content_length);
+			wsi->u.http.content_remain = wsi->u.http.content_length;
+		} else /* can't do 1.1 without a content length or chunked */
+			if (!wsi->chunked)
+				wsi->u.http.connection_type = HTTP_CONNECTION_CLOSE;
+
+		/*
+		 * we seem to be good to go, give client last chance to check
+		 * headers and OK it
+		 */
+		if (wsi->protocol->callback(wsi, LWS_CALLBACK_CLIENT_FILTER_PRE_ESTABLISH,
+					    wsi->user_space, NULL, 0))
+			goto bail2;
+
+		/* clear his proxy connection timeout */
+		lws_set_timeout(wsi, NO_PENDING_TIMEOUT, 0);
+
+		wsi->rxflow_change_to = LWS_RXFLOW_ALLOW;
+
+		/* call him back to inform him he is up */
+		if (wsi->protocol->callback(wsi,
+					    LWS_CALLBACK_ESTABLISHED_CLIENT_HTTP,
+					    wsi->user_space, NULL, 0))
+			goto bail3;
+
+		/* free up his parsing allocations */
+		lws_header_table_detach(wsi, 0);
+
+		lwsl_notice("%s: client connection up\n", __func__);
+
+		return 0;
+	}
+
 	if (lws_hdr_total_length(wsi, WSI_TOKEN_ACCEPT) == 0) {
 		lwsl_info("no ACCEPT\n");
 		isErrorCodeReceived = 1;
@@ -592,7 +565,7 @@ lws_client_interpret_server_handshake(struct lws *wsi)
 		 * no protocol name to work from,
 		 * default to first protocol
 		 */
-		wsi->protocol = &context->protocols[0];
+		wsi->protocol = &wsi->vhost->protocols[0];
 		goto check_extensions;
 	}
 
@@ -621,9 +594,9 @@ lws_client_interpret_server_handshake(struct lws *wsi)
 	 */
 	n = 0;
 	wsi->protocol = NULL;
-	while (context->protocols[n].callback && !wsi->protocol) {
-		if (strcmp(p, context->protocols[n].name) == 0) {
-			wsi->protocol = &context->protocols[n];
+	while (wsi->vhost->protocols[n].callback && !wsi->protocol) {
+		if (strcmp(p, wsi->vhost->protocols[n].name) == 0) {
+			wsi->protocol = &wsi->vhost->protocols[n];
 			break;
 		}
 		n++;
@@ -691,7 +664,7 @@ check_extensions:
 		lwsl_notice("checking client ext %s\n", ext_name);
 
 		n = 0;
-		ext = lws_get_context(wsi)->extensions;
+		ext = wsi->vhost->extensions;
 		while (ext && ext->callback) {
 			if (strcmp(ext_name, ext->name)) {
 				ext++;
@@ -802,7 +775,7 @@ check_accept:
 	lws_set_timeout(wsi, NO_PENDING_TIMEOUT, 0);
 
 	/* free up his parsing allocations */
-	lws_header_table_detach(wsi);
+	lws_header_table_detach(wsi, 0);
 
 	lws_union_transition(wsi, LWSCM_WS_CLIENT);
 	wsi->state = LWSS_ESTABLISHED;
@@ -844,7 +817,7 @@ check_accept:
 	 * inform all extensions, not just active ones since they
 	 * already know
 	 */
-	ext = context->extensions;
+	ext = wsi->vhost->extensions;
 
 	while (ext && ext->callback) {
 		v = NULL;
@@ -891,39 +864,36 @@ lws_generate_client_handshake(struct lws *wsi, char *pkt)
 {
 	char buf[128], hash[20], key_b64[40], *p = pkt;
 	struct lws_context *context = wsi->context;
+	const char *meth;
 	int n;
 #ifndef LWS_NO_EXTENSIONS
 	const struct lws_extension *ext;
 	int ext_count = 0;
 #endif
 
-	/*
-	 * create the random key
-	 */
-	n = lws_get_random(context, hash, 16);
-	if (n != 16) {
-		lwsl_err("Unable to read from random dev %s\n",
-			 SYSTEM_RANDOM_FILEPATH);
-		lws_close_free_wsi(wsi, LWS_CLOSE_STATUS_NOSTATUS);
-		return NULL;
+	meth = lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_METHOD);
+	if (!meth) {
+		meth = "GET";
+		wsi->do_ws = 1;
+	} else
+		wsi->do_ws = 0;
+
+	if (wsi->do_ws) {
+		/*
+		 * create the random key
+		 */
+		n = lws_get_random(context, hash, 16);
+		if (n != 16) {
+			lwsl_err("Unable to read from random dev %s\n",
+				 SYSTEM_RANDOM_FILEPATH);
+			lws_close_free_wsi(wsi, LWS_CLOSE_STATUS_NOSTATUS);
+			return NULL;
+		}
+
+		lws_b64_encode_string(hash, 16, key_b64, sizeof(key_b64));
 	}
 
-	lws_b64_encode_string(hash, 16, key_b64, sizeof(key_b64));
-
 	/*
-	 * 00 example client handshake
-	 *
-	 * GET /socket.io/websocket HTTP/1.1
-	 * Upgrade: WebSocket
-	 * Connection: Upgrade
-	 * Host: 127.0.0.1:9999
-	 * Origin: http://127.0.0.1
-	 * Sec-WebSocket-Key1: 1 0 2#0W 9 89 7  92 ^
-	 * Sec-WebSocket-Key2: 7 7Y 4328 B2v[8(z1
-	 * Cookie: socketio=websocket
-	 *
-	 * (Á®Ä0¶†≥
-	 *
 	 * 04 example client handshake
 	 *
 	 * GET /chat HTTP/1.1
@@ -936,7 +906,7 @@ lws_generate_client_handshake(struct lws *wsi, char *pkt)
 	 * Sec-WebSocket-Version: 4
 	 */
 
-	p += sprintf(p, "GET %s HTTP/1.1\x0d\x0a",
+	p += sprintf(p, "%s %s HTTP/1.1\x0d\x0a", meth,
 		     lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_URI));
 
 	p += sprintf(p, "Pragma: no-cache\x0d\x0a"
@@ -944,87 +914,89 @@ lws_generate_client_handshake(struct lws *wsi, char *pkt)
 
 	p += sprintf(p, "Host: %s\x0d\x0a",
 		     lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_HOST));
-	p += sprintf(p, "Upgrade: websocket\x0d\x0a"
-			"Connection: Upgrade\x0d\x0a"
-			"Sec-WebSocket-Key: ");
-	strcpy(p, key_b64);
-	p += strlen(key_b64);
-	p += sprintf(p, "\x0d\x0a");
+
 	if (lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_ORIGIN))
 		p += sprintf(p, "Origin: http://%s\x0d\x0a",
 			     lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_ORIGIN));
 
-	if (lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_SENT_PROTOCOLS))
-		p += sprintf(p, "Sec-WebSocket-Protocol: %s\x0d\x0a",
-		     lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_SENT_PROTOCOLS));
+	if (wsi->do_ws) {
+		p += sprintf(p, "Upgrade: websocket\x0d\x0a"
+				"Connection: Upgrade\x0d\x0a"
+				"Sec-WebSocket-Key: ");
+		strcpy(p, key_b64);
+		p += strlen(key_b64);
+		p += sprintf(p, "\x0d\x0a");
+		if (lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_SENT_PROTOCOLS))
+			p += sprintf(p, "Sec-WebSocket-Protocol: %s\x0d\x0a",
+			     lws_hdr_simple_ptr(wsi, _WSI_TOKEN_CLIENT_SENT_PROTOCOLS));
 
-	/* tell the server what extensions we could support */
+		/* tell the server what extensions we could support */
 
 #ifndef LWS_NO_EXTENSIONS
-	ext = context->extensions;
-	while (ext && ext->callback) {
-		n = lws_ext_cb_all_exts(context, wsi,
-			   LWS_EXT_CB_CHECK_OK_TO_PROPOSE_EXTENSION,
-			   (char *)ext->name, 0);
-		if (n) { /* an extension vetos us */
-			lwsl_ext("ext %s vetoed\n", (char *)ext->name);
+		ext = wsi->vhost->extensions;
+		while (ext && ext->callback) {
+			n = lws_ext_cb_all_exts(context, wsi,
+				   LWS_EXT_CB_CHECK_OK_TO_PROPOSE_EXTENSION,
+				   (char *)ext->name, 0);
+			if (n) { /* an extension vetos us */
+				lwsl_ext("ext %s vetoed\n", (char *)ext->name);
+				ext++;
+				continue;
+			}
+			n = wsi->vhost->protocols[0].callback(wsi,
+				LWS_CALLBACK_CLIENT_CONFIRM_EXTENSION_SUPPORTED,
+					wsi->user_space, (char *)ext->name, 0);
+
+			/*
+			 * zero return from callback means
+			 * go ahead and allow the extension,
+			 * it's what we get if the callback is
+			 * unhandled
+			 */
+
+			if (n) {
+				ext++;
+				continue;
+			}
+
+			/* apply it */
+
+			if (ext_count)
+				*p++ = ',';
+			else
+				p += sprintf(p, "Sec-WebSocket-Extensions: ");
+			p += sprintf(p, "%s", ext->client_offer);
+			ext_count++;
+
 			ext++;
-			continue;
 		}
-		n = context->protocols[0].callback(wsi,
-			LWS_CALLBACK_CLIENT_CONFIRM_EXTENSION_SUPPORTED,
-				wsi->user_space, (char *)ext->name, 0);
-
-		/*
-		 * zero return from callback means
-		 * go ahead and allow the extension,
-		 * it's what we get if the callback is
-		 * unhandled
-		 */
-
-		if (n) {
-			ext++;
-			continue;
-		}
-
-		/* apply it */
-
 		if (ext_count)
-			*p++ = ',';
-                else
-                        p += sprintf(p, "Sec-WebSocket-Extensions: ");
-
-		p += sprintf(p, "%s", ext->client_offer);
-		ext_count++;
-
-		ext++;
-	}
-	if (ext_count)
-		p += sprintf(p, "\x0d\x0a");
+			p += sprintf(p, "\x0d\x0a");
 #endif
 
-	if (wsi->ietf_spec_revision)
-		p += sprintf(p, "Sec-WebSocket-Version: %d\x0d\x0a",
-			     wsi->ietf_spec_revision);
+		if (wsi->ietf_spec_revision)
+			p += sprintf(p, "Sec-WebSocket-Version: %d\x0d\x0a",
+				     wsi->ietf_spec_revision);
+
+		/* prepare the expected server accept response */
+
+		key_b64[39] = '\0'; /* enforce composed length below buf sizeof */
+		n = sprintf(buf, "%s258EAFA5-E914-47DA-95CA-C5AB0DC85B11", key_b64);
+
+		lws_SHA1((unsigned char *)buf, n, (unsigned char *)hash);
+
+		lws_b64_encode_string(hash, 20,
+				      wsi->u.hdr.ah->initial_handshake_hash_base64,
+				      sizeof(wsi->u.hdr.ah->initial_handshake_hash_base64));
+	}
 
 	/* give userland a chance to append, eg, cookies */
 
-	context->protocols[0].callback(wsi, LWS_CALLBACK_CLIENT_APPEND_HANDSHAKE_HEADER,
-				       NULL, &p, (pkt + LWS_MAX_SOCKET_IO_BUF) - p - 12);
+	wsi->vhost->protocols[0].callback(wsi, LWS_CALLBACK_CLIENT_APPEND_HANDSHAKE_HEADER,
+				wsi->user_space, &p, (pkt + LWS_MAX_SOCKET_IO_BUF) - p - 12);
 
 	p += sprintf(p, "\x0d\x0a");
 
-	/* prepare the expected server accept response */
-
-	key_b64[39] = '\0'; /* enforce composed length below buf sizeof */
-	n = sprintf(buf, "%s258EAFA5-E914-47DA-95CA-C5AB0DC85B11", key_b64);
-
-	lws_SHA1((unsigned char *)buf, n, (unsigned char *)hash);
-
-	lws_b64_encode_string(hash, 20,
-			      wsi->u.hdr.ah->initial_handshake_hash_base64,
-			      sizeof(wsi->u.hdr.ah->initial_handshake_hash_base64));
-
 	return p;
 }
 
