@@ -22,6 +22,11 @@
 #include "lws_config.h"
 #include "lws_config_private.h"
 
+
+#if defined(LWS_WITH_CGI) && defined(LWS_HAVE_VFORK)
+#define  _GNU_SOURCE
+#endif
+
 #ifdef LWS_HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -77,12 +82,13 @@
 #include <in6addr.h>
 #endif
 #include <mstcpip.h>
+#include <io.h>
 
 #ifndef __func__
 #define __func__ __FUNCTION__
 #endif
 
-#ifdef _WIN32_WCE
+#if (defined(_MSC_VER) && _MSC_VER < 1900) || defined(_WIN32_WCE)
 #define vsnprintf _vsnprintf
 #else
 #ifdef LWS_HAVE__VSNPRINTF
@@ -107,6 +113,10 @@
 #include <netdb.h>
 #include <signal.h>
 #include <sys/socket.h>
+#ifdef LWS_WITH_HTTP_PROXY
+#include <hubbub/hubbub.h>
+#include <hubbub/parser.h>
+#endif
 #ifdef LWS_BUILTIN_GETIFADDRS
  #include <getifaddrs.h>
 #else
@@ -114,6 +124,7 @@
 #endif
 #if defined (__ANDROID__)
 #include <syslog.h>
+#include <sys/resource.h>
 #else
 #include <sys/syslog.h>
 #endif
@@ -152,7 +163,7 @@
 #define LWS_POLLHUP (POLLHUP|POLLERR)
 #define LWS_POLLIN (POLLIN)
 #define LWS_POLLOUT (POLLOUT)
-#define compatible_close(fd) close(fd)
+static inline int compatible_close(int fd) { return close(fd); }
 #define lws_set_blocking_send(wsi)
 
 #ifdef MBED_OPERATORS
@@ -175,6 +186,7 @@
 #endif
 
 #ifdef LWS_OPENSSL_SUPPORT
+
 #ifdef USE_WOLFSSL
 #ifdef USE_OLD_CYASSL
 #include <cyassl/openssl/ssl.h>
@@ -184,11 +196,33 @@
 #include <wolfssl/error-ssl.h>
 #endif /* not USE_OLD_CYASSL */
 #else
+#if defined(LWS_USE_POLARSSL)
+#include <polarssl/ssl.h>
+#include <polarssl/error.h>
+#include <polarssl/md5.h>
+#include <polarssl/sha1.h>
+#include <polarssl/ecdh.h>
+#define SSL_ERROR_WANT_READ POLARSSL_ERR_NET_WANT_READ
+#define SSL_ERROR_WANT_WRITE POLARSSL_ERR_NET_WANT_WRITE
+#define OPENSSL_VERSION_NUMBER  0x10002000L
+#else
+#if defined(LWS_USE_MBEDTLS)
+#include <mbedtls/ssl.h>
+#include <mbedtls/error.h>
+#include <mbedtls/md5.h>
+#include <mbedtls/sha1.h>
+#include <mbedtls/ecdh.h>
+#else
 #include <openssl/ssl.h>
 #include <openssl/evp.h>
 #include <openssl/err.h>
 #include <openssl/md5.h>
 #include <openssl/sha.h>
+#ifdef LWS_HAVE_OPENSSL_ECDH_H
+#include <openssl/ecdh.h>
+#endif
+#endif /* not USE_MBEDTLS */
+#endif /* not USE_POLARSSL */
 #endif /* not USE_WOLFSSL */
 #endif
 
@@ -248,11 +282,6 @@ typedef unsigned __int64 u_int64_t;
 #ifdef __cplusplus
 extern "C" {
 #endif
-#include <stddef.h>
-
-#ifndef container_of
-#define container_of(P,T,M)	((T *)((char *)(P) - offsetof(T, M)))
-#endif
 
 #if defined(__QNX__)
 	#include <gulliver.h>
@@ -351,6 +380,7 @@ enum lws_connection_states {
 	LWSS_HTTP_BODY,
 	LWSS_DEAD_SOCKET,
 	LWSS_ESTABLISHED,
+	LWSS_CLIENT_HTTP_ESTABLISHED,
 	LWSS_CLIENT_UNCONNECTED,
 	LWSS_RETURNED_CLOSE_ALREADY,
 	LWSS_AWAITING_CLOSE_ACK,
@@ -360,11 +390,14 @@ enum lws_connection_states {
 	LWSS_HTTP2_AWAIT_CLIENT_PREFACE,
 	LWSS_HTTP2_ESTABLISHED_PRE_SETTINGS,
 	LWSS_HTTP2_ESTABLISHED,
+
+	LWSS_CGI,
 };
 
 enum http_version {
 	HTTP_VERSION_1_0,
 	HTTP_VERSION_1_1,
+	HTTP_VERSION_2
 };
 
 enum http_connection_type {
@@ -410,7 +443,9 @@ enum lws_rx_parse_state {
 
 enum connection_mode {
 	LWSCM_HTTP_SERVING,
+	LWSCM_HTTP_CLIENT, /* we are client to someone else's server */
 	LWSCM_HTTP_SERVING_ACCEPTED, /* actual HTTP service going on */
+	LWSCM_HTTP_CLIENT_ACCEPTED, /* actual HTTP service going on */
 	LWSCM_PRE_WS_SERVING_ACCEPT,
 
 	LWSCM_WS_SERVING,
@@ -434,6 +469,7 @@ enum connection_mode {
 
 	/* special internal types */
 	LWSCM_SERVER_LISTENER,
+	LWSCM_CGI, /* stdin, stdout, stderr for another cgi master wsi */
 };
 
 enum {
@@ -516,7 +552,6 @@ struct allocated_headers {
 
 #ifndef LWS_NO_CLIENT
 	char initial_handshake_hash_base64[30];
-	unsigned short c_port;
 #endif
 
 	unsigned short pos;
@@ -538,6 +573,9 @@ struct lws_context_per_thread {
 	struct lws *tx_draining_ext_list;
 	struct lws *timeout_list;
 	struct lws_context *context;
+#ifdef LWS_WITH_CGI
+	struct lws_cgi *cgi_list;
+#endif
 	void *http_header_data;
 	struct allocated_headers *ah_pool;
 	struct lws *ah_wait_list;
@@ -555,6 +593,7 @@ struct lws_context_per_thread {
 	uv_loop_t *io_loop_uv;
 	uv_signal_t signals[8];
 	uv_timer_t uv_timeout_watcher;
+	uv_idle_t uv_idle;
 #endif
 #if defined(LWS_USE_LIBEV)
 	struct lws_io_watcher w_accept;
@@ -563,7 +602,6 @@ struct lws_context_per_thread {
 	struct lws_signal_watcher w_sigint;
 	unsigned char ev_loop_foreign:1;
 #endif
-	lws_sockfd_type lserv_fd;
 
 	unsigned long count_conns;
 	/*
@@ -583,17 +621,77 @@ struct lws_context_per_thread {
 	unsigned char tid;
 };
 
+/*
+ * virtual host -related context information
+ *   vhostwide SSL context
+ *   vhostwide proxy
+ *
+ * heirarchy:
+ *
+ * context -> vhost -> wsi
+ *
+ * incoming connection non-SSL vhost binding:
+ *
+ *    listen socket -> wsi -> select vhost after first headers
+ *
+ * incoming connection SSL vhost binding:
+ *
+ *    SSL SNI -> wsi -> bind after SSL negotiation
+ */
+
+struct lws_vhost {
+	char http_proxy_address[128];
+	char proxy_basic_auth_token[128];
+	struct lws_context *context;
+	struct lws_vhost *vhost_next;
+	const struct lws_http_mount *mount_list;
+	struct lws *lserv_wsi;
+	const char *name;
+	const char *iface;
+	const struct lws_protocols *protocols;
+	void **protocol_vh_privs;
+	const struct lws_protocol_vhost_options *pvo;
+	struct lws **same_vh_protocol_list;
+#ifdef LWS_OPENSSL_SUPPORT
+	SSL_CTX *ssl_ctx;
+	SSL_CTX *ssl_client_ctx;
+#endif
+#ifndef LWS_NO_EXTENSIONS
+	const struct lws_extension *extensions;
+#endif
+	unsigned long long rx, tx;
+	unsigned long conn, trans, ws_upgrades, http2_upgrades;
+
+	int listen_port;
+	unsigned int http_proxy_port;
+	unsigned int options;
+	int count_protocols;
+	int ka_time;
+	int ka_probes;
+	int ka_interval;
+	int keepalive_timeout;
+#ifdef LWS_WITH_ACCESS_LOG
+	int log_fd;
+#endif
+
+#ifdef LWS_OPENSSL_SUPPORT
+	int use_ssl;
+	int allow_non_ssl_on_ssl_port;
+	unsigned int user_supplied_ssl_ctx:1;
+#endif
+	unsigned char default_protocol_index;
+};
+
 /*
  * the rest is managed per-context, that includes
  *
  *  - processwide single fd -> wsi lookup
  *  - contextwide headers pool
- *  - contextwide ssl context
- *  - contextwide proxy
  */
 
 struct lws_context {
 	time_t last_timeout_check_s;
+	time_t time_up;
 	struct lws_plat_file_ops fops;
 	struct lws_context_per_thread pt[LWS_MAX_SMP];
 #ifdef _WIN32
@@ -602,27 +700,18 @@ struct lws_context {
 #else
 	struct lws **lws_lookup;  /* fd to wsi */
 #endif
-	const char *iface;
+	struct lws_vhost *vhost_list;
+	struct lws_plugin *plugin_list;
 	const struct lws_token_limits *token_limits;
 	void *user_space;
+	const char *server_string;
 
-	const struct lws_protocols *protocols;
-
-#ifdef LWS_OPENSSL_SUPPORT
-	SSL_CTX *ssl_ctx;
-	SSL_CTX *ssl_client_ctx;
-#endif
-#ifndef LWS_NO_EXTENSIONS
-	const struct lws_extension *extensions;
-#endif
 #if defined(LWS_USE_LIBEV)
 	lws_ev_signal_cb_t * lws_ev_sigint_cb;
 #endif
 #if defined(LWS_USE_LIBUV)
-	lws_uv_signal_cb_t * lws_uv_sigint_cb;
+	uv_signal_cb lws_uv_sigint_cb;
 #endif
-	char http_proxy_address[128];
-	char proxy_basic_auth_token[128];
 	char canonical_hostname[128];
 #ifdef LWS_LATENCY
 	unsigned long worst_latency;
@@ -630,16 +719,26 @@ struct lws_context {
 #endif
 
 	int max_fds;
-	int listen_port;
 #if defined(LWS_USE_LIBEV) || defined(LWS_USE_LIBUV)
 	int use_ev_sigint;
 #endif
 	int started_with_parent;
+	int uid, gid;
 
 	int fd_random;
-	int lserv_mod;
+#ifdef LWS_OPENSSL_SUPPORT
+#define lws_ssl_anybody_has_buffered_read(w) \
+		(w->vhost->use_ssl && \
+		 w->context->pt[(int)w->tsi].pending_read_list)
+#define lws_ssl_anybody_has_buffered_read_tsi(c, t) \
+		(/*c->use_ssl && */ \
+		 c->pt[(int)t].pending_read_list)
+#else
+#define lws_ssl_anybody_has_buffered_read(ctx) (0)
+#define lws_ssl_anybody_has_buffered_read_tsi(ctx, t) (0)
+#endif
 	int count_wsi_allocated;
-	unsigned int http_proxy_port;
+	int count_cgi_spawned;
 	unsigned int options;
 	unsigned int fd_limit_per_thread;
 	unsigned int timeout_secs;
@@ -653,39 +752,32 @@ struct lws_context {
 	volatile int service_tid;
 	int service_tid_detected;
 
-	int count_protocols;
-	int ka_time;
-	int ka_probes;
-	int ka_interval;
-
-#ifdef LWS_OPENSSL_SUPPORT
-	int use_ssl;
-	int allow_non_ssl_on_ssl_port;
-	unsigned int user_supplied_ssl_ctx:1;
-#define lws_ssl_anybody_has_buffered_read(w) \
-		(w->context->use_ssl && \
-		 w->context->pt[(int)w->tsi].pending_read_list)
-#define lws_ssl_anybody_has_buffered_read_tsi(c, t) \
-		(c->use_ssl && \
-		 c->pt[(int)t].pending_read_list)
-#else
-#define lws_ssl_anybody_has_buffered_read(ctx) (0)
-#define lws_ssl_anybody_has_buffered_read_tsi(ctx, t) (0)
-#endif
-
 	short max_http_header_data;
 	short max_http_header_pool;
 	short count_threads;
+	short plugin_protocol_count;
+	short plugin_extension_count;
+	short server_string_len;
 
 	unsigned int being_destroyed:1;
 	unsigned int requested_kill:1;
+	unsigned int protocol_init_done:1;
 };
 
+#define lws_get_context_protocol(ctx, x) ctx->vhost_list->protocols[x]
+#define lws_get_vh_protocol(vh, x) vh->protocols[x]
+
 LWS_EXTERN void
 lws_close_free_wsi_final(struct lws *wsi);
 LWS_EXTERN void
 lws_libuv_closehandle(struct lws *wsi);
 
+LWS_VISIBLE LWS_EXTERN int
+lws_plat_plugins_init(struct lws_context * context, const char * const *d);
+
+LWS_VISIBLE LWS_EXTERN int
+lws_plat_plugins_destroy(struct lws_context * context);
+
 enum {
 	LWS_EV_READ = (1 << 0),
 	LWS_EV_WRITE = (1 << 1),
@@ -706,7 +798,7 @@ LWS_EXTERN void
 lws_libev_destroyloop(struct lws_context *context, int tsi);
 LWS_EXTERN void
 lws_libev_run(const struct lws_context *context, int tsi);
-#define LWS_LIBEV_ENABLED(context) (context->options & LWS_SERVER_OPTION_LIBEV)
+#define LWS_LIBEV_ENABLED(context) lws_check_opt(context->options, LWS_SERVER_OPTION_LIBEV)
 LWS_EXTERN void lws_feature_status_libev(struct lws_context_creation_info *info);
 #else
 #define lws_libev_accept(_a, _b) ((void) 0)
@@ -734,7 +826,7 @@ LWS_EXTERN void
 lws_libuv_run(const struct lws_context *context, int tsi);
 LWS_EXTERN void
 lws_libuv_destroyloop(struct lws_context *context, int tsi);
-#define LWS_LIBUV_ENABLED(context) (context->options & LWS_SERVER_OPTION_LIBUV)
+#define LWS_LIBUV_ENABLED(context) lws_check_opt(context->options, LWS_SERVER_OPTION_LIBUV)
 LWS_EXTERN void lws_feature_status_libuv(struct lws_context_creation_info *info);
 #else
 #define lws_libuv_accept(_a, _b) ((void) 0)
@@ -754,11 +846,17 @@ LWS_EXTERN void lws_feature_status_libuv(struct lws_context_creation_info *info)
 
 #ifdef LWS_USE_IPV6
 #define LWS_IPV6_ENABLED(context) \
-	(!(context->options & LWS_SERVER_OPTION_DISABLE_IPV6))
+	(!lws_check_opt(context->options, LWS_SERVER_OPTION_DISABLE_IPV6))
 #else
 #define LWS_IPV6_ENABLED(context) (0)
 #endif
 
+#ifdef LWS_USE_UNIX_SOCK
+#define LWS_UNIX_SOCK_ENABLED(vhost) \
+	(vhost->options & LWS_SERVER_OPTION_UNIX_SOCK)
+#else
+#define LWS_UNIX_SOCK_ENABLED(vhost) (0)
+#endif
 enum uri_path_states {
 	URIPS_IDLE,
 	URIPS_SEEN_SLASH,
@@ -784,10 +882,49 @@ enum uri_esc_states {
  * used interchangeably to access the same data
  */
 
+
+#ifndef LWS_NO_CLIENT
+struct client_info_stash {
+	char address[256];
+	char path[1024];
+	char host[256];
+	char origin[256];
+	char protocol[256];
+	char method[16];
+};
+#endif
+
+struct _lws_header_related {
+	/* MUST be first in struct */
+	struct allocated_headers *ah;
+	struct lws *ah_wait_list;
+	unsigned char *preamble_rx;
+#ifndef LWS_NO_CLIENT
+	struct client_info_stash *stash;
+#endif
+	unsigned int preamble_rx_len;
+	enum uri_path_states ups;
+	enum uri_esc_states ues;
+	short lextable_pos;
+	unsigned short current_token_limit;
+#ifndef LWS_NO_CLIENT
+	unsigned short c_port;
+#endif
+	char esc_stash;
+	char post_literal_equal;
+	unsigned char parser_state; /* enum lws_token_indexes */
+	char redirects;
+};
+
 struct _lws_http_mode_related {
 	/* MUST be first in struct */
 	struct allocated_headers *ah; /* mirroring  _lws_header_related */
 	struct lws *ah_wait_list;
+	unsigned char *preamble_rx;
+#ifndef LWS_NO_CLIENT
+	struct client_info_stash *stash;
+#endif
+	unsigned int preamble_rx_len;
 	struct lws *new_wsi_list;
 	unsigned long filepos;
 	unsigned long filelen;
@@ -946,20 +1083,6 @@ struct _lws_http2_related {
 
 #endif
 
-struct _lws_header_related {
-	/* MUST be first in struct */
-	struct allocated_headers *ah;
-	struct lws *ah_wait_list;
-	enum uri_path_states ups;
-	enum uri_esc_states ues;
-	short lextable_pos;
-	unsigned short current_token_limit;
-	char esc_stash;
-	char post_literal_equal;
-	unsigned char parser_state; /* enum lws_token_indexes */
-	char redirects;
-};
-
 struct _lws_websocket_related {
 	/* cheapest way to deal with ah overlap with ws union transition */
 	struct _lws_header_related hdr;
@@ -1002,6 +1125,46 @@ struct _lws_websocket_related {
 	unsigned int tx_draining_ext:1;
 };
 
+#ifdef LWS_WITH_CGI
+
+/* wsi who is master of the cgi points to an lws_cgi */
+
+struct lws_cgi {
+	struct lws_cgi *cgi_list;
+	struct lws *stdwsi[3]; /* points to the associated stdin/out/err wsis */
+	struct lws *wsi; /* owner */
+	unsigned long content_length;
+	unsigned long content_length_seen;
+	int pipe_fds[3][2];
+	int pid;
+
+	unsigned int being_closed:1;
+};
+#endif
+
+signed char char_to_hex(const char c);
+
+#ifndef LWS_NO_CLIENT
+enum lws_chunk_parser {
+	ELCP_HEX,
+	ELCP_CR,
+	ELCP_CONTENT,
+	ELCP_POST_CR,
+	ELCP_POST_LF,
+};
+#endif
+
+struct lws_rewrite;
+
+#ifdef LWS_WITH_ACCESS_LOG
+struct lws_access_log {
+	char *header_log;
+	char *user_agent;
+	unsigned long sent;
+	int response;
+};
+#endif
+
 struct lws {
 
 	/* structs */
@@ -1029,9 +1192,20 @@ struct lws {
 	/* pointers */
 
 	struct lws_context *context;
+	struct lws_vhost *vhost;
+	struct lws *parent; /* points to parent, if any */
+	struct lws *child_list; /* points to first child */
+	struct lws *sibling_list; /* subsequent children at same level */
+#ifdef LWS_WITH_CGI
+	struct lws_cgi *cgi; /* wsi being cgi master have one of these */
+#endif
 	const struct lws_protocols *protocol;
+	struct lws **same_vh_protocol_prev, *same_vh_protocol_next;
 	struct lws *timeout_list;
 	struct lws **timeout_list_prev;
+#ifdef LWS_WITH_ACCESS_LOG
+	struct lws_access_log access_log;
+#endif
 	void *user_space;
 	/* rxflow handling */
 	unsigned char *rxflow_buffer;
@@ -1043,9 +1217,14 @@ struct lws {
 #endif
 #ifdef LWS_OPENSSL_SUPPORT
 	SSL *ssl;
+#if !defined(LWS_USE_POLARSSL) && !defined(LWS_USE_MBEDTLS)
 	BIO *client_bio;
+#endif
 	struct lws *pending_read_list_prev, *pending_read_list_next;
 #endif
+#ifdef LWS_WITH_HTTP_PROXY
+	struct lws_rewrite *rw;
+#endif
 #ifdef LWS_LATENCY
 	unsigned long action_start;
 	unsigned long latency_start;
@@ -1060,12 +1239,34 @@ struct lws {
 	unsigned int trunc_alloc_len; /* size of malloc */
 	unsigned int trunc_offset; /* where we are in terms of spilling */
 	unsigned int trunc_len; /* how much is buffered */
+#ifndef LWS_NO_CLIENT
+	int chunk_remaining;
+#endif
+	unsigned int cache_secs;
 
 	unsigned int hdr_parsing_completed:1;
+	unsigned int http2_substream:1;
+	unsigned int listener:1;
 	unsigned int user_space_externally_allocated:1;
 	unsigned int socket_is_permanently_unusable:1;
 	unsigned int rxflow_change_to:2;
 	unsigned int more_rx_waiting:1; /* has to live here since ah may stick to end */
+	unsigned int conn_stat_done:1;
+	unsigned int cache_reuse:1;
+	unsigned int cache_revalidate:1;
+	unsigned int cache_intermediaries:1;
+	unsigned int favoured_pollin:1;
+#ifdef LWS_WITH_ACCESS_LOG
+	unsigned int access_log_pending:1;
+#endif
+#ifndef LWS_NO_CLIENT
+	unsigned int do_ws:1; /* whether we are doing http or ws flow */
+	unsigned int chunked:1; /* if the clientside connection is chunked */
+	unsigned int client_rx_avail:1;
+#endif
+#ifdef LWS_WITH_HTTP_PROXY
+	unsigned int perform_rewrite:1;
+#endif
 #ifndef LWS_NO_EXTENSIONS
 	unsigned int extension_data_pending:1;
 #endif
@@ -1076,6 +1277,9 @@ struct lws {
 #ifdef _WIN32
 	unsigned int sock_send_blocking:1;
 #endif
+#ifdef LWS_OPENSSL_SUPPORT
+	unsigned int redirect_to_https:1;
+#endif
 
 	/* chars */
 #ifndef LWS_NO_EXTENSIONS
@@ -1090,10 +1294,21 @@ struct lws {
 	char pending_timeout; /* enum pending_timeout */
 	char pps; /* enum lws_pending_protocol_send */
 	char tsi; /* thread service index we belong to */
+#ifdef LWS_WITH_CGI
+	char cgi_channel; /* which of stdin/out/err */
+	char hdr_state;
+#endif
+#ifndef LWS_NO_CLIENT
+	char chunk_parser; /* enum lws_chunk_parser */
+#endif
 };
 
 LWS_EXTERN int log_level;
 
+LWS_EXTERN int
+lws_socket_bind(struct lws_vhost *vhost, int sockfd, int port,
+		const char *iface);
+
 LWS_EXTERN void
 lws_close_free_wsi(struct lws *wsi, enum lws_close_status);
 
@@ -1176,7 +1391,7 @@ lws_client_reset(struct lws *wsi, int ssl, const char *address, int port,
 		 const char *path, const char *host);
 
 LWS_EXTERN struct lws * LWS_WARN_UNUSED_RESULT
-lws_create_new_server_wsi(struct lws_context *context);
+lws_create_new_server_wsi(struct lws_vhost *vhost);
 
 LWS_EXTERN char * LWS_WARN_UNUSED_RESULT
 lws_generate_client_handshake(struct lws *wsi, char *pkt);
@@ -1184,6 +1399,9 @@ lws_generate_client_handshake(struct lws *wsi, char *pkt);
 LWS_EXTERN int
 lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd);
 
+LWS_EXTERN struct lws *
+lws_client_connect_via_info2(struct lws *wsi);
+
 /*
  * EXTENSIONS
  */
@@ -1268,16 +1486,16 @@ void lws_http2_configure_if_upgraded(struct lws *wsi);
 #endif
 
 LWS_EXTERN int
-lws_plat_set_socket_options(struct lws_context *context, lws_sockfd_type fd);
+lws_plat_set_socket_options(struct lws_vhost *vhost, lws_sockfd_type fd);
 
 LWS_EXTERN int LWS_WARN_UNUSED_RESULT
-lws_header_table_attach(struct lws *wsi);
+lws_header_table_attach(struct lws *wsi, int autoservice);
 
 LWS_EXTERN int
-lws_header_table_detach(struct lws *wsi);
+lws_header_table_detach(struct lws *wsi, int autoservice);
 
 LWS_EXTERN void
-lws_header_table_reset(struct lws *wsi);
+lws_header_table_reset(struct lws *wsi, int autoservice);
 
 LWS_EXTERN char * LWS_WARN_UNUSED_RESULT
 lws_hdr_simple_ptr(struct lws *wsi, enum lws_token_indexes h);
@@ -1285,7 +1503,7 @@ lws_hdr_simple_ptr(struct lws *wsi, enum lws_token_indexes h);
 LWS_EXTERN int LWS_WARN_UNUSED_RESULT
 lws_hdr_simple_create(struct lws *wsi, enum lws_token_indexes h, const char *s);
 
-LWS_EXTERN int
+LWS_EXTERN int LWS_WARN_UNUSED_RESULT
 lws_ensure_user_space(struct lws *wsi);
 
 LWS_EXTERN int
@@ -1293,7 +1511,9 @@ lws_change_pollfd(struct lws *wsi, int _and, int _or);
 
 #ifndef LWS_NO_SERVER
 int lws_context_init_server(struct lws_context_creation_info *info,
-			    struct lws_context *context);
+			    struct lws_vhost *vhost);
+LWS_EXTERN struct lws_vhost *
+lws_select_vhost(struct lws_context *context, int port, const char *servername);
 LWS_EXTERN int
 handshake_0405(struct lws_context *context, struct lws *wsi);
 LWS_EXTERN int LWS_WARN_UNUSED_RESULT
@@ -1336,7 +1556,9 @@ enum lws_ssl_capable_status {
 #define lws_server_socket_service_ssl(_b, _c) (0)
 #define lws_ssl_close(_a) (0)
 #define lws_ssl_context_destroy(_a)
+#define lws_ssl_SSL_CTX_destroy(_a)
 #define lws_ssl_remove_wsi_from_buffered_list(_a)
+#define lws_context_init_ssl_library(_a)
 #else
 #define LWS_SSL_ENABLED(context) (context->use_ssl)
 LWS_EXTERN int openssl_websocket_private_data_index;
@@ -1346,29 +1568,41 @@ LWS_EXTERN int LWS_WARN_UNUSED_RESULT
 lws_ssl_capable_write(struct lws *wsi, unsigned char *buf, int len);
 LWS_EXTERN int LWS_WARN_UNUSED_RESULT
 lws_ssl_pending(struct lws *wsi);
+LWS_EXTERN int
+lws_context_init_ssl_library(struct lws_context_creation_info *info);
 LWS_EXTERN int LWS_WARN_UNUSED_RESULT
 lws_server_socket_service_ssl(struct lws *new_wsi, lws_sockfd_type accept_fd);
 LWS_EXTERN int
 lws_ssl_close(struct lws *wsi);
 LWS_EXTERN void
+lws_ssl_SSL_CTX_destroy(struct lws_vhost *vhost);
+LWS_EXTERN void
 lws_ssl_context_destroy(struct lws_context *context);
 LWS_VISIBLE void
 lws_ssl_remove_wsi_from_buffered_list(struct lws *wsi);
+LWS_EXTERN int
+lws_ssl_client_bio_create(struct lws *wsi);
+LWS_EXTERN int
+lws_ssl_client_connect1(struct lws *wsi);
+LWS_EXTERN int
+lws_ssl_client_connect2(struct lws *wsi);
+LWS_EXTERN void
+lws_ssl_elaborate_error(void);
 #ifndef LWS_NO_SERVER
 LWS_EXTERN int
 lws_context_init_server_ssl(struct lws_context_creation_info *info,
-		 	    struct lws_context *context);
+			    struct lws_vhost *vhost);
 #else
 #define lws_context_init_server_ssl(_a, _b) (0)
 #endif
 LWS_EXTERN void
-lws_ssl_destroy(struct lws_context *context);
+lws_ssl_destroy(struct lws_vhost *vhost);
 
 /* HTTP2-related */
 
 #ifdef LWS_USE_HTTP2
 LWS_EXTERN void
-lws_context_init_http2_ssl(struct lws_context *context);
+lws_context_init_http2_ssl(struct lws_vhost *vhost);
 #else
 #define lws_context_init_http2_ssl(_a)
 #endif
@@ -1380,6 +1614,13 @@ lws_pt_mutex_init(struct lws_context_per_thread *pt)
 {
 	pthread_mutex_init(&pt->lock, NULL);
 }
+
+static LWS_INLINE void
+lws_pt_mutex_destroy(struct lws_context_per_thread *pt)
+{
+	pthread_mutex_destroy(&pt->lock);
+}
+
 static LWS_INLINE void
 lws_pt_lock(struct lws_context_per_thread *pt)
 {
@@ -1393,6 +1634,7 @@ lws_pt_unlock(struct lws_context_per_thread *pt)
 }
 #else
 #define lws_pt_mutex_init(_a) (void)(_a)
+#define lws_pt_mutex_destroy(_a) (void)(_a)
 #define lws_pt_lock(_a) (void)(_a)
 #define lws_pt_unlock(_a) (void)(_a)
 #endif
@@ -1406,14 +1648,41 @@ lws_ssl_capable_write_no_ssl(struct lws *wsi, unsigned char *buf, int len);
 LWS_EXTERN int LWS_WARN_UNUSED_RESULT
 lws_ssl_pending_no_ssl(struct lws *wsi);
 
+#ifdef LWS_WITH_HTTP_PROXY
+struct lws_rewrite {
+	hubbub_parser *parser;
+	hubbub_parser_optparams params;
+	const char *from, *to;
+	int from_len, to_len;
+	unsigned char *p, *end;
+	struct lws *wsi;
+};
+static LWS_INLINE int hstrcmp(hubbub_string *s, const char *p, int len)
+{
+	if (s->len != len)
+		return 1;
+
+	return strncmp((const char *)s->ptr, p, len);
+}
+typedef hubbub_error (*hubbub_callback_t)(const hubbub_token *token, void *pw);
+LWS_EXTERN struct lws_rewrite *
+lws_rewrite_create(struct lws *wsi, hubbub_callback_t cb, const char *from, const char *to);
+LWS_EXTERN void
+lws_rewrite_destroy(struct lws_rewrite *r);
+LWS_EXTERN int
+lws_rewrite_parse(struct lws_rewrite *r, const unsigned char *in, int in_len);
+#endif
+
 #ifndef LWS_NO_CLIENT
 LWS_EXTERN int lws_client_socket_service(struct lws_context *context,
 					 struct lws *wsi,
 					 struct lws_pollfd *pollfd);
+LWS_EXTERN int LWS_WARN_UNUSED_RESULT
+lws_http_transaction_completed_client(struct lws *wsi);
 #ifdef LWS_OPENSSL_SUPPORT
 LWS_EXTERN int
 lws_context_init_client_ssl(struct lws_context_creation_info *info,
-			    struct lws_context *context);
+			    struct lws_vhost *vhost);
 #else
 	#define lws_context_init_client_ssl(_a, _b) (0)
 #endif
@@ -1450,6 +1719,22 @@ LWS_EXTERN int
 lws_get_addresses(struct lws_context *context, void *ads, char *name,
 		  int name_len, char *rip, int rip_len);
 
+LWS_EXTERN const char *
+lws_get_peer_simple(struct lws *wsi, char *name, int namelen);
+
+#ifdef LWS_WITH_ACCESS_LOG
+LWS_EXTERN int
+lws_access_log(struct lws *wsi);
+#else
+#define lws_access_log(_a)
+#endif
+
+LWS_EXTERN int
+lws_cgi_kill_terminated(struct lws_context_per_thread *pt);
+
+int
+lws_protocol_init(struct lws_context *context);
+
 /*
  * custom allocator
  */
